/*
Copyright 2015 by Joseph Forgione
This file is part of VCC (Virtual Color Computer).

    VCC (Virtual Color Computer) is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    VCC (Virtual Color Computer) is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VCC (Virtual Color Computer).  If not, see <http://www.gnu.org/licenses/>.
*/

#include "stdio.h"
#include "defines.h"
#include "iobus.h"
#include "mc6821.h"
#include "pakinterface.h"
#include "tcc1014registers.h"
#include "tcc1014mmu.h"
#include "logger.h"
#include "config.h"
unsigned char port_read(unsigned short addr)
{
	unsigned char port=0,temp=0;
//	char Message[128]="";
	port = (addr & 0xFF);
/*
if ( (port>=0x50) & (port <=0x5a))
{
		sprintf(Message,"Reading from %x\n",addr);
		WriteLog(Message,TOCONS);
}
*/
	switch (port)
	{

		case 0:
		case 1:
		case 2:
		case 3:
			temp=pia0_read(port);	//MC6821 P.I.A  Keyboard access $FF00-$FF03
		break;

		case 0x20:
		case 0x21:
		case 0x22:
		case 0x23:

			temp=pia1_read(port);	//MC6821 P.I.A	Sound and VDG Control 
		break;

		case 0xC0:
		case 0xC1:
		case 0xC2:
		case 0xC3:
		case 0xC4:
		case 0xC5:
		case 0xC6:
		case 0xC7:
		case 0xC8:
		case 0xC9:
		case 0xCA:
		case 0xCB:
		case 0xCC:
		case 0xCD:
		case 0xCE:
		case 0xCF:
		case 0xD0:
		case 0xD1:
		case 0xD2:
		case 0xD3:
		case 0xD4:
		case 0xD5:
		case 0xD6:
		case 0xD7:
		case 0xD8:
		case 0xD9:
		case 0xDA:
		case 0xDB:
		case 0xDC:
		case 0xDD:
		case 0xDE:
		case 0xDF:
			temp=sam_read(port);	//MC6883 S.A.M. address range $FFC0-$FFDF
		break;

		case 0xF0:
		case 0xF1:
		case 0xF2:
		case 0xF3:
		case 0xF4:
		case 0xF5:
		case 0xF6:
		case 0xF7:
		case 0xF8:
		case 0xF9:
		case 0xFA:
		case 0xFB:
		case 0xFC:
		case 0xFD:
		case 0xFE:
		case 0xFF:
			temp=sam_read(port);	// SAM controls IRQ Vectors at $FFF0 - $FFFF	
		break;

		case 0x90:					//TCC1014 G.I.M.E.
		case 0x91:
		case 0x92:
		case 0x93:
		case 0x94:
		case 0x95:
		case 0x96:
		case 0x97:
		case 0x98:
		case 0x99:
		case 0x9A:
		case 0x9B:
		case 0x9C:
		case 0x9D:
		case 0x9E:
		case 0x9F:
		case 0xA0:
		case 0xA1:
		case 0xA2:
		case 0xA3:
		case 0xA4:
		case 0xA5:
		case 0xA6:
		case 0xA7:
		case 0xA8:
		case 0xA9:
		case 0xAA:
		case 0xAB:
		case 0xAC:
		case 0xAD:
		case 0xAE:
		case 0xAF:
		case 0xB0:
		case 0xB1:
		case 0xB2:
		case 0xB3:
		case 0xB4:
		case 0xB5:
		case 0xB6:
		case 0xB7:
		case 0xB8:
		case 0xB9:
		case 0xBA:
		case 0xBB:
		case 0xBC:
		case 0xBD:
		case 0xBE:
		case 0xBF:
			temp=GimeRead(port);
		break;
		default:
			temp=PackPortRead (port);
		}
	return(temp);
}

extern void DumpRegisters(void);
static int debug = 0;

void port_write(unsigned char data,unsigned short addr)
{
	unsigned char port=0;
	port = (addr & 0xFF);
	/*
	char Message[256]="";
	if ( (addr>=0x50) & (port <=0x5a))
	{
			sprintf(Message,"Wrting %x to %x\n",data,addr);
			WriteLog(Message,TOCONS);
	}
	*/
	switch (port)
	{

		case 0:
		case 1:
		case 2:
		case 3:
			pia0_write(data,port);	//MC6821 P.I.A  Keyboard access $FF00-$FF03
		break;

		case 0x20:
		case 0x21:
		case 0x22:
		case 0x23:
			pia1_write(data,port);	//MC6821 P.I.A	Sound and VDG Control 
		break;
#define DEBUGPORT 1
#ifdef  DEBUGPORT
		case 0x70:
		    if (debug) fputc((int)data, stderr);
			break;

		case 0x71:
		    if (debug) fprintf(stderr, "%02X", (int)data);
			break;

		case 0x72:
		    if (debug) fprintf(stderr, "%02X", (int)data);
			break;

		case 0x73:
		    if (debug) fputc('\n', stderr);
			break;

		case 0x74:
			debug = (int)data;
			break;

		case 0x75:
			DumpRegisters();
			break;
#endif
		case 0xC0:
		case 0xC1:
		case 0xC2:
		case 0xC3:
		case 0xC4:
		case 0xC5:
		case 0xC6:
		case 0xC7:
		case 0xC8:
		case 0xC9:
		case 0xCA:
		case 0xCB:
		case 0xCC:
		case 0xCD:
		case 0xCE:
		case 0xCF:
		case 0xD0:
		case 0xD1:
		case 0xD2:
		case 0xD3:
		case 0xD4:
		case 0xD5:
		case 0xD6:
		case 0xD7:
		case 0xD8:
		case 0xD9:
		case 0xDA:
		case 0xDB:
		case 0xDC:
		case 0xDD:
		case 0xDE:
		case 0xDF:
			sam_write(data,port);	//MC6883 S.A.M. address range $FFC0-$FFDF
		break;

		case 0x90:
		case 0x91:
		case 0x92:
		case 0x93:
		case 0x94:
		case 0x95:
		case 0x96:
		case 0x97:
		case 0x98:
		case 0x99:
		case 0x9A:
		case 0x9B:
		case 0x9C:
		case 0x9D:
		case 0x9E:
		case 0x9F:
		case 0xA0:
		case 0xA1:
		case 0xA2:
		case 0xA3:
		case 0xA4:
		case 0xA5:
		case 0xA6:
		case 0xA7:
		case 0xA8:
		case 0xA9:
		case 0xAA:
		case 0xAB:
		case 0xAC:
		case 0xAD:
		case 0xAE:
		case 0xAF:
		case 0xB0:
		case 0xB1:
		case 0xB2:
		case 0xB3:
		case 0xB4:
		case 0xB5:
		case 0xB6:
		case 0xB7:
		case 0xB8:
		case 0xB9:
		case 0xBA:
		case 0xBB:
		case 0xBC:
		case 0xBD:
		case 0xBE:
		case 0xBF:
			GimeWrite(port,data);
		break;
		default:
			PackPortWrite (port,data);
	}
	return;
}

